// This may look like C code, but it is really -*- C++ -*-
//
// history:
// - mComplex renamed to sComplex, lComplex added Apr13
// - mComplex added on Apr99 for large systems to save memory 
// ToDo:
// - add fix point number x*2^(-15) as typedef, + put, get
// - class sComplex => typedef sComplex + put, get
//

#ifndef _Complex_h
#ifdef __GNUG__
/* #pragma interface */
#endif
#define _Complex_h 1


// #include <iostream.h>
#include <math.h>


class sComplex;		/* declaration Apr99 */

class  Complex
{
#ifdef __ATT_complex__
public:
#else
protected:
#endif

  double           re;
  double           im;

public:

  double           real() const;
  double           imag() const;

                   Complex();
                   Complex(const  Complex& y);
                   Complex(const sComplex& y);
                   Complex(double r, double i=0);

                  ~Complex();

  Complex&         operator =  (const Complex& y);
  Complex&         operator =  (double y);               /* JS */

  Complex&         operator += (const Complex& y);
  Complex&         operator += (double y);
  Complex&         operator -= (const Complex& y);
  Complex&         operator -= (double y);
  Complex&         operator *= (const Complex& y);
  Complex&         operator *= (double y);

  Complex&         operator /= (const Complex& y); 
  Complex&         operator /= (double y); 

  void             error(const char* msg) const;
};

/* ====================================================================
   same again with float instead of double (templates are more risky?)
   ==================================================================== */
class sComplex
{
#ifdef __ATT_complex__
public:
#else
protected:
#endif

  sdouble           re;
  sdouble           im;

public:

  sdouble           mreal() const;
  sdouble           mimag() const;

                   sComplex();
                   sComplex(const sComplex& y);
                   sComplex(const  Complex& y);
                   sComplex(sdouble r, sdouble i=0);

                  ~sComplex();

  sComplex&        operator += (const sComplex& y);
  sComplex&        operator -= (const sComplex& y);
  sComplex&        operator *= (const sComplex& y);

  void             error(const char* msg) const;
};


inline sdouble  sComplex::mreal() const { return re; }
inline sdouble  sComplex::mimag() const { return im; }

/* ====================================================================
   same again with long double instead of double (templates are more risky?)
   ==================================================================== */
class lComplex
{
#ifdef __ATT_complex__
public:
#else
protected:
#endif

  ldouble           re;
  ldouble           im;

public:

  ldouble           mreal() const;
  ldouble           mimag() const;

                   lComplex();
                   lComplex(const lComplex& y);
                   lComplex(const  Complex& y);
                   lComplex(ldouble r, ldouble i=0);

                  ~lComplex();

  lComplex&        operator += (const lComplex& y);
  lComplex&        operator -= (const lComplex& y);
  lComplex&        operator *= (const lComplex& y);

  void             error(const char* msg) const;
};


inline ldouble  lComplex::mreal() const { return re; }
inline ldouble  lComplex::mimag() const { return im; }

// non-inline functions

#if 0
Complex   operator /  (const Complex& x, const Complex& y);
Complex   operator /  (const Complex& x, double y);
Complex   operator /  (double   x, const Complex& y);


Complex   cos(const Complex& x);
Complex   sin(const Complex& x);

Complex   cosh(const Complex& x);
Complex   sinh(const Complex& x);

Complex   exp(const Complex& x);
Complex   log(const Complex& x);

Complex   pow(const Complex& x, int p);
Complex   pow(const Complex& x, const Complex& p);
Complex   pow(const Complex& x, double y);
Complex   sqrt(const Complex& x);

istream&  operator >> (istream& s, Complex& x);
ostream&  operator << (ostream& s, const Complex& x);

// other functions defined as inlines

int  operator == (const Complex& x, const Complex& y);
int  operator == (const Complex& x, double y);
int  operator != (const Complex& x, const Complex& y);
int  operator != (const Complex& x, double y);

Complex  operator - (const Complex& x);
Complex  conj(const Complex& x);
Complex  operator + (const Complex& x, const Complex& y);
Complex  operator + (const Complex& x, double y);
Complex  operator + (double x, const Complex& y);
Complex  operator - (const Complex& x, const Complex& y);
Complex  operator - (const Complex& x, double y);
Complex  operator - (double x, const Complex& y);
Complex  operator * (const Complex& x, const Complex& y);
Complex  operator * (const Complex& x, double y);
Complex  operator * (double x, const Complex& y);

double  real(const Complex& x);
/* double  real(const double& x); */    /* JS */
double  imag(const Complex& x);
double  abs(const Complex& x);
double  fabs(const Complex& x); /* JS */
double  norm(const Complex& x);
double  arg(const Complex& x);

Complex polar(double r, double t = 0.0);

#endif
// inline members

inline double  Complex::real() const { return re; }
inline double  Complex::imag() const { return im; }

inline Complex::Complex() {}
inline Complex::Complex(const  Complex& y) :re(y.real()), im(y.imag()) {}
inline Complex::Complex(const sComplex& y) :re(y.mreal()), im(y.mimag()) {}
inline Complex::Complex(double r, double i) :re(r), im(i) {}

inline Complex::~Complex() {}

inline Complex&  Complex::operator =  (const Complex& y) 
{ re = y.real(); im = y.imag(); return *this;
} 

           /* JS */
inline Complex&  Complex::operator =  (double y) 
{ re = y; im = 0.0; return *this; } 


inline Complex&  Complex::operator += (const Complex& y)
{ re += y.real();  im += y.imag(); return *this;
}

inline Complex&  Complex::operator += (double y)
{ re += y; return *this;
}

inline Complex&  Complex::operator -= (const Complex& y)
{ re -= y.real();  im -= y.imag(); return *this;
}

inline Complex&  Complex::operator -= (double y)
{ re -= y; return *this;
}

inline Complex&  Complex::operator *= (const Complex& y)
{ double r = re * y.real() - im * y.imag();
  im = re * y.imag() + im * y.real(); 
  re = r; 
  return *this; 
}

inline Complex&  Complex::operator *= (double y)
{ re *=  y; im *=  y; return *this;}

inline Complex&  Complex::operator /= (const Complex& y) /* JS */
{ double r=(re*y.real()+im*y.imag())/(y.real()*y.real()+y.imag()*y.imag());
  im =  (-re*y.imag()+im*y.real())/(y.real()*y.real()+y.imag()*y.imag());
  re = r; 
  return *this; 
}

inline Complex&  Complex::operator /= (double y)         /* JS */
{ re /=  y; im /=  y; return *this;}

//  functions

inline int  operator == (const Complex& x, const Complex& y)
{ return x.real() == y.real() && x.imag() == y.imag();}

inline int  operator == (const Complex& x, double y)
{ return x.imag() == 0.0 && x.real() == y;}

inline int  operator != (const Complex& x, const Complex& y)
{ return x.real() != y.real() || x.imag() != y.imag();}

inline int  operator != (const Complex& x, double y)
{ return x.imag() != 0.0 || x.real() != y;}

inline Complex  operator - (const Complex& x)
{ return Complex(-x.real(), -x.imag());}

inline Complex  conj(const Complex& x)
{ return Complex(x.real(), -x.imag());}

inline Complex  operator + (const Complex& x, const Complex& y)
{ return Complex(x.real() + y.real(), x.imag() + y.imag());}

inline Complex  operator + (const Complex& x, double y)
{ return Complex(x.real() + y, x.imag());}

inline Complex  operator + (double x, const Complex& y)
{ return Complex(x + y.real(), y.imag());}

inline Complex  operator - (const Complex& x, const Complex& y)
{ return Complex(x.real() - y.real(), x.imag() - y.imag());}

inline Complex  operator - (const Complex& x, double y)
{ return Complex(x.real() - y, x.imag());}

inline Complex  operator - (double x, const Complex& y)
{ return Complex(x - y.real(), -y.imag());}

inline Complex  operator * (const Complex& x, const Complex& y)
{ return Complex(x.real() * y.real() - x.imag() * y.imag(),
                 x.real() * y.imag() + x.imag() * y.real());}

inline Complex  operator * (const Complex& x, double y)
{  return Complex(x.real() * y, x.imag() * y);}

inline Complex  operator * (double x, const Complex& y)
{ return Complex(x * y.real(), x * y.imag());}

inline Complex  operator / (const Complex& x, const Complex& y) /* JS */
{ return Complex(
  ( x.real()*y.real()+x.imag()*y.imag())/(y.real()*y.real()+y.imag()*y.imag()),
  (-x.real()*y.imag()+x.imag()*y.real())/(y.real()*y.real()+y.imag()*y.imag()));}

inline Complex  operator / (const Complex& x, double y) /* JS */
{  return Complex(x.real() / y, x.imag() / y);}

inline Complex  operator / (double x, const Complex& y) /* JS */
{ return Complex(
   x*y.real()/(y.real()*y.real()+y.imag()*y.imag()),
  -x*y.imag()/(y.real()*y.real()+y.imag()*y.imag()));}

// inline double  real(const double& x) /* JS */
// { return x;}

inline double  real(const Complex& x)
{ return x.real();}

inline double  imag(const Complex& x)
{ return x.imag();}

inline double  abs(const Complex& x)
{ return hypot(x.real(), x.imag());}

inline double  fabs(const Complex& x) /* JS zeitintensiv! */
{ return hypot(x.real(), x.imag());}

inline double  norm(const Complex& x)
{ return (x.real() * x.real() + x.imag() * x.imag());}

inline double  arg(const Complex& x)
{ return atan2(x.imag(), x.real());}

inline Complex  polar(double r, double t)
{ return Complex(r * cos(t), r * sin(t));}

inline Complex  sqrt(const Complex& x)  /* JS */
{ return Complex(sqrt(abs(x))*cos(arg(x)/2.),
                 sqrt(abs(x))*sin(arg(x)/2.));}

/* ====================================================================
   same again with float instead of double (templates are more risky?)
   ==================================================================== */

// inline members


inline sComplex::sComplex() {}
// following line was out-commented until v1.7.8b, why? error if VecType=4
inline sComplex::sComplex(const sComplex& y) :re(y.mreal()), im(y.mimag()) {}
inline sComplex::sComplex(const  Complex& y) :re(y.real()), im(y.imag()) {}
inline sComplex::sComplex(sdouble r, sdouble i) :re(r), im(i) {}

inline sComplex::~sComplex() {}

inline sComplex&  sComplex::operator += (const sComplex& y)
{ re += y.mreal();  im += y.mimag(); return *this;
}
inline sComplex&  sComplex::operator -= (const sComplex& y)
{ re -= y.mreal();  im -= y.mimag(); return *this;
}

inline sComplex&  sComplex::operator *= (const sComplex& y)
{ sdouble r = re * y.mreal() - im * y.mimag();
  im = re * y.mimag() + im * y.mreal(); 
  re = r; 
  return *this; 
}

/* ====================================================================
   same again with long double instead of double (templates are more risky?)
   ==================================================================== */

// inline members


inline lComplex::lComplex() {}
// following line was out-commented until v1.7.8b, why? error if VecType=4
inline lComplex::lComplex(const lComplex& y) :re(y.mreal()), im(y.mimag()) {}
inline lComplex::lComplex(const  Complex& y) :re(y.real()), im(y.imag()) {}
inline lComplex::lComplex(ldouble r, ldouble i) :re(r), im(i) {}

inline lComplex::~lComplex() {}

inline lComplex&  lComplex::operator += (const lComplex& y)
{ re += y.mreal();  im += y.mimag(); return *this;
}
inline lComplex&  lComplex::operator -= (const lComplex& y)
{ re -= y.mreal();  im -= y.mimag(); return *this;
}

inline lComplex&  lComplex::operator *= (const lComplex& y)
{ ldouble r = re * y.mreal() - im * y.mimag();
  im = re * y.mimag() + im * y.mreal(); 
  re = r; 
  return *this; 
}

#endif
